/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support  -  ROUSSET  -
 * ----------------------------------------------------------------------------
 * Copyright (c) 2008, Atmel Corporation

 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 * File Name           : pm9g45.c
 * Object              :
 * Creation            : 21.04.2010
 *-----------------------------------------------------------------------------
 */
#include "part.h"
#include "gpio.h"
#include "pmc.h"
#include "rstc.h"
#include "debug.h"
#include "main.h"
#include "ddramc.h"
#ifdef CONFIG_NANDFLASH
#include "nandflash.h"
#endif
#ifdef CONFIG_DATAFLASH
#include "dataflash.h"
#endif

#ifndef CONFIG_THUMB
static inline unsigned int get_cp15(void)
{
	unsigned int value;
	__asm__("mrc p15, 0, %0, c1, c0, 0" : "=r" (value));
	return value;
}

static inline void set_cp15(unsigned int value)
{
	__asm__("mcr p15, 0, %0, c1, c0, 0" : : "r" (value));
}
#else
int get_cp15(void);
void set_cp15(unsigned int value);
#endif

#ifdef CONFIG_HW_INIT

/*--------------------------------------------------------------------------*/
/* \fn    hw_get_revision							      					*/
/* \brief This function returns the hardware revision of the board 			*/
/* This function returns the hardware revision of the board	      			*/
/*--------------------------------------------------------------------------*/
int hw_get_revision(void)
{
int ret;

	ret = pio_get_value(CONFIG_VERSION_PIN1) << 1;
	ret |= pio_get_value(CONFIG_VERSION_PIN0);
	return (~ret) & 3;
}

/*----------------------------------------------------------------------------*/
/* \fn    hw_init							      */
/* \brief This function performs very low level HW initialization	      */
/* This function is invoked as soon as possible during the c_startup	      */
/* The bss segment must be initialized					      */
/*----------------------------------------------------------------------------*/
void hw_init(void)
{
	unsigned int cp15, reg, i;
	unsigned int reg_mckr, reg_pllar;

	/* Configure PIOs */
	const struct pio_desc hw_pio[] = {
#ifdef CONFIG_VERBOSE
		{"RXD", AT91C_PIN_PB(12), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"TXD", AT91C_PIN_PB(13), 0, PIO_DEFAULT, PIO_PERIPH_A},
#endif
		/* Hardware revision of the board */
		{"REV0", CONFIG_VERSION_PIN0, 0, PIO_PULLUP, PIO_INPUT},
		{"REV1", CONFIG_VERSION_PIN1, 0, PIO_PULLUP, PIO_INPUT},
		{(char *) 0, 0, 0, PIO_DEFAULT, PIO_PERIPH_A},
	};

	/* Disable watchdog */
	writel(AT91C_WDTC_WDDIS, AT91C_BASE_WDTC + WDTC_WDMR);


	// Do not initialize the PLLA if it is already
	// initialized by a JTAG emulator like PEEDI
	reg_mckr = pmc_cfg_read(PMC_MCKR);
	reg_pllar = pmc_cfg_read(PMC_PLLAR);
	if ( (reg_mckr != MCKR_CSS_SETTINGS) ||	(reg_pllar != PLLA_SETTINGS) )
	{
		reg_mckr = MCKR_CSS_SETTINGS;
		reg_pllar = PLLA_SETTINGS;

		/* At this stage the main oscillator is supposed to be enabled
		 * PCK = MCK = MOSC */
		writel(0x00, AT91C_BASE_PMC + PMC_PLLICPR);

		/* Configure PLLA = MOSC * (PLL_MULA + 1) / PLL_DIVA */
		pmc_cfg_plla(reg_pllar, PLL_LOCK_TIMEOUT);

		/* PCK = PLLA/2 = 3 * MCK */
		pmc_cfg_mck(BOARD_PRESCALER, PLL_LOCK_TIMEOUT);

		/* Switch MCK on PLLA output */
		pmc_cfg_mck(reg_mckr, PLL_LOCK_TIMEOUT);

		/* Configure PLLB/PLLUTMI in order to reset USB device UTMI */
		pmc_cfg_pllutmi(PLLUTMI_SETTINGS, PLL_LOCK_TIMEOUT);
	}

	/* Disable External Reset */
	writel(AT91C_RSTC_KEY_UNLOCK & ~AT91C_RSTC_URSTEN, AT91C_BASE_RSTC + RSTC_RMR);

	/* Configure CP15 */
	cp15 = get_cp15();
	cp15 |= I_CACHE;
	set_cp15(cp15);

	/* Configure the PIO controller */
	writel((1 << AT91C_ID_PIOA), (PMC_PCER + AT91C_BASE_PMC));
	pio_setup(hw_pio);

#ifdef CONFIG_VERBOSE
	/* Enable Debug messages on the DBGU */
	dbg_init(BAUDRATE(MASTER_CLOCK, 115200));

	dbg_print("\r\nStart AT91Bootstrap...\n\r");
#endif /* CONFIG_VERBOSE */

	/* 
	 * Slow clock source
	 * Switch to external 32.768KHz crystal resonator and disable
	 * internal RC oscillator if it is not.
	 * If 32.768KHz oscillator(NOT crystal resonator) is connected
	 * bypass the internal oscillator and reset and discard OSC332EN.
	 */
	reg = readl(AT91C_SYS_SLCKSEL);
	if (!((reg & AT91C_SLCKSEL_OSCSEL) && (reg & AT91C_SLCKSEL_OSC32EN))) {
#ifdef CONFIG_VERBOSE
		dbg_print("\r\n-I- Slow clock on internal RC ... switching to external crystal!\n\r");
#endif /* CONFIG_VERBOSE */
		/* Enable oscillator */
		writel(reg | AT91C_SLCKSEL_OSC32EN, AT91C_SYS_SLCKSEL);
		/* 
		 * Wait 32768 Hz Startup Time for clock stabilization
		 * (software loop)
		 * The worst case from "46.6 32 kHz Oscillator Characteristics"
		 * is 1200ms. 1000ms it works.
		 */
		for(i=0x1000000;i>0;i--)
			reg = readl(AT91C_SYS_SLCKSEL);
		/* Switch from internal to external clock */
		writel(reg | AT91C_SLCKSEL_OSCSEL, AT91C_SYS_SLCKSEL);
		/* Wait 5 slow clock cycles for internal resynchronization */
		reg = readl(AT91C_SYS_SLCKSEL);
		/* Disable the RC oscillator */
		/* if disabled Linux kernel halts with last message 
		 * "ehci_hcd: USB 2.0 'Enhanced' Host Controller (EHCI) Driver"
		 *
		 * writel(reg & ~AT91C_SLCKSEL_RCEN, AT91C_SYS_SLCKSEL);
		 */

	} else {
#ifdef CONFIG_VERBOSE
		dbg_print("\r\n-I- Slow clock on external crystal.\n\r");
#endif /* CONFIG_VERBOSE */
	}

#ifdef CONFIG_DDR2
	/* Configure DDRAM Controller */
	ddramc_hw_init();
#endif /* CONFIG_DDR2 */
}
#endif /* CONFIG_HW_INIT */

#ifdef CONFIG_DDR2
// 128MB, 2x8-bit MT47H64M8
static	SDdramConfig ddram_config_2x8bit_128MB = {
	.ddramc_mdr  = (AT91C_DDRC2_DBW_16_BITS | AT91C_DDRC2_MD_DDR2_SDRAM),
		
	.ddramc_cr   = (AT91C_DDRC2_NC_DDR10_SDR9  |     	// 10 column bits (1K)
			AT91C_DDRC2_NR_14          |     	// 14 row bits    (8K)
			AT91C_DDRC2_CAS_3          |     	// CAS Latency 3
			AT91C_DDRC2_DLL_RESET_DISABLED),	// DLL not reset

	.ddramc_cr2   = 0xFFFFFFFF, 				// not used
	.ddramc_rtr  = 	0x40F,

	.ddramc_t0pr = (AT91C_DDRC2_TRAS_6  |		//  6 * 7.5 = 45   ns
			AT91C_DDRC2_TRCD_2  |		//  2 * 7.5 = 22.5 ns
			AT91C_DDRC2_TWR_2   |		//  2 * 7.5 = 15   ns
			AT91C_DDRC2_TRC_8  |		//  8 * 7.5 = 75   ns
			AT91C_DDRC2_TRP_2   |		//  2 * 7.5 = 22.5 ns
			AT91C_DDRC2_TRRD_1  |		//  1 * 7.5 = 7.5   ns
			AT91C_DDRC2_TWTR_1  |		//  1 clock cycle
			AT91C_DDRC2_TMRD_2),		//  2 clock cycles
		
	.ddramc_t1pr = (AT91C_DDRC2_TXP_2  |		//  2 * 7.5 = 15 ns
			200 << 16           |		// 200 clock cycles, TXSRD: Exit self refresh delay to Read command
			16 << 8             |		// 16 * 7.5 = 120 ns TXSNR: Exit self refresh delay to non read command
			AT91C_DDRC2_TRFC_14 << 0),	// 14 * 7.5 = 142 ns (must be 140 ns for 1Gb DDR)

	.ddramc_t2pr = (AT91C_DDRC2_TRTP_1   |		//  1 * 7.5 = 7.5 ns
			AT91C_DDRC2_TRPA_0   |		//  0 * 7.5 = 0 ns
			AT91C_DDRC2_TXARDS_7 |		//  7 clock cycles
			AT91C_DDRC2_TXARD_2),		//  2 clock cycles
};

// 64MB, 1x16-bit MT47H32M16
static	SDdramConfig ddram_config_1x16bit_64MB =
{
	.ddramc_mdr  = (AT91C_DDRC2_DBW_16_BITS | AT91C_DDRC2_MD_DDR2_SDRAM),

	// initialize the DDR2 controller with this value in
	// order to ensure proper initialization of the DDR2 chip
	//========================================================
	.ddramc_cr   = (AT91C_DDRC2_NC_DDR10_SDR9  |     	// 10 column bits (1K)
					AT91C_DDRC2_NR_14          |     	// 14 row bits    (8K)
					AT91C_DDRC2_CAS_3          |     	// CAS Latency 3
					AT91C_DDRC2_DLL_RESET_DISABLED),	// DLL not reset

	// initialize the DDR2 controller with this
	// value after the DDR2 chip initialization is done
	//=================================================
	.ddramc_cr2   = (AT91C_DDRC2_NC_DDR10_SDR9  |     	// 10 column bits (1K)
			AT91C_DDRC2_NR_13          |     	// 13 row bits    (4K)
			AT91C_DDRC2_CAS_3          |     	// CAS Latency 3
			AT91C_DDRC2_DLL_RESET_DISABLED),	// DLL not reset

	.ddramc_rtr  = 	0x40F,

	.ddramc_t0pr = (AT91C_DDRC2_TRAS_6  |		//  6 * 7.5 = 45   ns
			AT91C_DDRC2_TRCD_2  |		//  2 * 7.5 = 22.5 ns
			AT91C_DDRC2_TWR_2   |		//  2 * 7.5 = 15   ns
			AT91C_DDRC2_TRC_8  |		//  8 * 7.5 = 75   ns
			AT91C_DDRC2_TRP_2   |		//  2 * 7.5 = 22.5 ns
			AT91C_DDRC2_TRRD_1  |		//  1 * 7.5 = 7.5   ns
			AT91C_DDRC2_TWTR_1  |		//  1 clock cycle
			AT91C_DDRC2_TMRD_2),		//  2 clock cycles

	.ddramc_t1pr = (AT91C_DDRC2_TXP_2  |		//  2 * 7.5 = 15 ns
			200 << 16           |		// 200 clock cycles, TXSRD: Exit self refresh delay to Read command
			16 << 8             |		// 16 * 7.5 = 120 ns TXSNR: Exit self refresh delay to non read command
			AT91C_DDRC2_TRFC_14 << 0),	// 14 * 7.5 = 142 ns (must be 140 ns for 1Gb DDR)

	.ddramc_t2pr = (AT91C_DDRC2_TRTP_1   |		//  1 * 7.5 = 7.5 ns
			AT91C_DDRC2_TRPA_0   |		//  0 * 7.5 = 0 ns
			AT91C_DDRC2_TXARDS_7 |		//  7 clock cycles
			AT91C_DDRC2_TXARD_2),		//  2 clock cycles
};

/*------------------------------------------------------------------------------*/
/* \fn    ddramc_hw_init							*/
/* \brief This function performs DDRAMC HW initialization			*/
/*------------------------------------------------------------------------------*/
void ddramc_hw_init(void)
{
int revision;
unsigned int reg;

	// Get board revision:
	// v1.1 (read 0) - 128MB at 0x20000000, 2x8-bit MT47H64M8
	//		   			64MB at 0x70000000, 1x16-bit MT47H32M16
	// v1.2 (read 1) - 128MB at 0x70000000, 2x8bit MT47H64M8
	//========================================================
	revision = hw_get_revision();

	// ENABLE DDR2 clock 
	writel(AT91C_PMC_DDR, AT91C_BASE_PMC + PMC_SCER);

	reg = readl(AT91C_BASE_CCFG + CCFG_EBICSA);

	/* EBI Chip Select 1 is assigned to the SDRAM Controller. */
	reg |= AT91C_EBI_CS1A_SDRAMC;

	/* optimized for 1.8V powered memories with High Drive */
	reg &= ~(3 << 16);

	/* High Drive data bus for stand-alone DDR controller */
	reg &= ~(1 << 18);
	
	/* 0 = EBI D0 - D15 Data Bus bits are internally pulled-up to the VDDIOM1 power supply.
	 * 1 = EBI D0 - D15 Data Bus bits are not internally pulled-up. */
	reg |=  AT91C_EBI_DBPUC;

	writel(reg, AT91C_BASE_CCFG + CCFG_EBICSA);

	switch( revision )
	{
		// board revision 1.0 and 1.1
		//===========================
		case 0:
			/* DDRAM2 controller, 0x70000000 */
			ddram_init(AT91C_BASE_DDR2C, AT91C_DDR2, &ddram_config_1x16bit_64MB);

			/* EBI DDRAM controller, 0x20000000 */
			ddram_init(AT91C_BASE_DDR2CP1, AT91C_EBI_CS1, &ddram_config_2x8bit_128MB);
			break;

		// board revision v1.2 and later
		//==============================
		default:
			/* DDRAM2 controller, 0x70000000 */
			ddram_init(AT91C_BASE_DDR2C, AT91C_DDR2, &ddram_config_2x8bit_128MB);
			break;
	}
}
#endif /* CONFIG_DDR2 */

#ifdef CONFIG_DATAFLASH
#if	defined(CONFIG_DATAFLASH_RECOVERY)
/*------------------------------------------------------------------------------*/
/* \fn    df_recovery								*/
/* \brief This function erases DataFlash Page 0 if BP4 is pressed 		*/
/*        during boot sequence							*/
/*------------------------------------------------------------------------------*/
void df_recovery(AT91PS_DF pDf)
{

}
#endif
/*------------------------------------------------------------------------------*/
/* \fn    df_hw_init								*/
/* \brief This function performs DataFlash HW initialization			*/
/*------------------------------------------------------------------------------*/
void df_hw_init(void)
{
	/* Configure PIOs */
	const struct pio_desc df_pio[] = {
		{"MISO",  AT91C_PIN_PB(0), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"MOSI",  AT91C_PIN_PB(1), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"SPCK",  AT91C_PIN_PB(2), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"NPCS0", AT91C_PIN_PB(3), 0, PIO_DEFAULT, PIO_PERIPH_A},
		{"NPCS1", AT91C_PIN_PC(11), 0, PIO_DEFAULT, PIO_PERIPH_B},
		{(char *) 0, 0, 0, PIO_DEFAULT, PIO_PERIPH_A},
	};

	/* Configure the PIO controller */
	pio_setup(df_pio);
}
#endif /* CONFIG_DATAFLASH */



#ifdef CONFIG_NANDFLASH
/*------------------------------------------------------------------------------*/
/* \fn    nand_recovery						*/
/* \brief This function erases NandFlash Block 0 if BP4 is pressed 		*/
/*        during boot sequence							*/
/*------------------------------------------------------------------------------*/
static void nand_recovery(void)
{

}

/*------------------------------------------------------------------------------*/
/* \fn    nandflash_hw_init							*/
/* \brief NandFlash HW init							*/
/*------------------------------------------------------------------------------*/
void nandflash_hw_init(void)
{
	/* Configure PIOs */
	const struct pio_desc nand_pio[] = {
		{"RDY_BSY", AT91C_PIN_PD(3), 0, PIO_PULLUP, PIO_INPUT},
		{"NANDCS",  AT91C_PIN_PC(14), 0, PIO_PULLUP, PIO_OUTPUT},
		{(char *) 0, 0, 0,  PIO_DEFAULT, PIO_PERIPH_A},
	};

	/* Setup Smart Media, first enable the address range of CS3 in HMATRIX user interface */
	writel(readl(AT91C_BASE_CCFG + CCFG_EBICSA) | AT91C_EBI_CS3A_SM, AT91C_BASE_CCFG + CCFG_EBICSA);
	
	/* EBI IO in 1.8V mode */
	writel(readl(AT91C_BASE_CCFG + CCFG_EBICSA) & ~(1<<16), AT91C_BASE_CCFG + CCFG_EBICSA);
	 
	/* Configure SMC CS3 */
 	writel((AT91C_SM_NWE_SETUP | AT91C_SM_NCS_WR_SETUP | AT91C_SM_NRD_SETUP | AT91C_SM_NCS_RD_SETUP), AT91C_BASE_SMC + SMC_SETUP3);
  	writel((AT91C_SM_NWE_PULSE | AT91C_SM_NCS_WR_PULSE | AT91C_SM_NRD_PULSE | AT91C_SM_NCS_RD_PULSE), AT91C_BASE_SMC + SMC_PULSE3);
	writel((AT91C_SM_NWE_CYCLE | AT91C_SM_NRD_CYCLE)						, AT91C_BASE_SMC + SMC_CYCLE3);
	writel((AT91C_SMC_READMODE | AT91C_SMC_WRITEMODE | AT91C_SMC_NWAITM_NWAIT_DISABLE | 
  		AT91C_SMC_DBW_WIDTH_SIXTEEN_BITS | AT91C_SM_TDF)						, AT91C_BASE_SMC + SMC_CTRL3);

	/* Configure the PIO controller */
	writel((1 << AT91C_ID_PIOC), (PMC_PCER + AT91C_BASE_PMC));
	writel((1 << AT91C_ID_PIOD_E), (PMC_PCER + AT91C_BASE_PMC));
	pio_setup(nand_pio);
	
	nand_recovery();
}

/*------------------------------------------------------------------------------*/
/* \fn    nandflash_cfg_16bits_dbw_init						*/
/* \brief Configure SMC in 16 bits mode						*/
/*------------------------------------------------------------------------------*/
void nandflash_cfg_16bits_dbw_init(void)
{
	writel(readl(AT91C_BASE_SMC + SMC_CTRL3) | AT91C_SMC_DBW_WIDTH_SIXTEEN_BITS, AT91C_BASE_SMC + SMC_CTRL3);
}

/*------------------------------------------------------------------------------*/
/* \fn    nandflash_cfg_8bits_dbw_init						*/
/* \brief Configure SMC in 8 bits mode						*/
/*------------------------------------------------------------------------------*/
void nandflash_cfg_8bits_dbw_init(void)
{
	writel((readl(AT91C_BASE_SMC + SMC_CTRL3) & ~(AT91C_SMC_DBW)) | AT91C_SMC_DBW_WIDTH_EIGTH_BITS, AT91C_BASE_SMC + SMC_CTRL3);
}

#endif /* #ifdef CONFIG_NANDFLASH */
